"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const converter_1 = require("../../converter");
var CommandEntryPoint;
(function (CommandEntryPoint) {
    CommandEntryPoint[CommandEntryPoint["CellContextMenu"] = 0] = "CellContextMenu";
    CommandEntryPoint[CommandEntryPoint["FileEditorContextMenu"] = 1] = "FileEditorContextMenu";
})(CommandEntryPoint = exports.CommandEntryPoint || (exports.CommandEntryPoint = {}));
function toDocumentChanges(changes) {
    let documentChanges = [];
    for (let uri of Object.keys(changes)) {
        documentChanges.push({
            textDocument: { uri },
            edits: changes[uri]
        });
    }
    return documentChanges;
}
class CodeMirrorLSPFeature {
    constructor(virtual_editor, virtual_document, connection, jupyterlab_components) {
        this.virtual_editor = virtual_editor;
        this.virtual_document = virtual_document;
        this.connection = connection;
        this.jupyterlab_components = jupyterlab_components;
        this.editor_handlers = new Map();
        this.connection_handlers = new Map();
        this.wrapper_handlers = new Map();
        this.is_registered = false;
    }
    register() {
        // register editor handlers
        for (let [event_name, handler] of this.editor_handlers) {
            this.virtual_editor.on(event_name, handler);
        }
        // register connection handlers
        for (let [event_name, handler] of this.connection_handlers) {
            this.connection.on(event_name, handler);
        }
        // register editor wrapper handlers
        this.wrapper = this.virtual_editor.getWrapperElement();
        for (let [event_name, handler] of this.wrapper_handlers) {
            this.wrapper.addEventListener(event_name, handler);
        }
        this.is_registered = true;
    }
    isEnabled() {
        // TODO - use settings
        return true;
    }
    remove() {
        // unregister editor handlers
        for (let [event_name, handler] of this.editor_handlers) {
            this.virtual_editor.off(event_name, handler);
        }
        // unregister connection handlers
        for (let [event_name, handler] of this.connection_handlers) {
            this.connection.off(event_name, handler);
        }
        // unregister editor wrapper handlers
        for (let [event_name, handler] of this.wrapper_handlers) {
            this.wrapper.removeEventListener(event_name, handler);
        }
    }
    afterChange(change, root_position) {
        // nothing here, yet
    }
    range_to_editor_range(range, cm_editor) {
        let start = converter_1.PositionConverter.lsp_to_cm(range.start);
        let end = converter_1.PositionConverter.lsp_to_cm(range.end);
        if (typeof cm_editor === 'undefined') {
            let start_in_root = this.transform_virtual_position_to_root_position(start);
            cm_editor = this.virtual_editor.get_editor_at_root_position(start_in_root);
        }
        return {
            start: this.virtual_document.transform_virtual_to_editor(start),
            end: this.virtual_document.transform_virtual_to_editor(end),
            editor: cm_editor
        };
    }
    position_from_mouse(event) {
        return this.virtual_editor.coordsChar({
            left: event.clientX,
            top: event.clientY
        }, 'window');
    }
    transform_virtual_position_to_root_position(start) {
        let cm_editor = this.virtual_document.virtual_lines.get(start.line).editor;
        let editor_position = this.virtual_document.transform_virtual_to_editor(start);
        return this.virtual_editor.transform_editor_to_root(cm_editor, editor_position);
    }
    get_cm_editor(position) {
        return this.virtual_editor.get_cm_editor(position);
    }
    get_language_at(position, editor) {
        return editor.getModeAt(position).name;
    }
    extract_last_character(change) {
        if (change.origin === 'paste') {
            return change.text[0][change.text.length - 1];
        }
        else {
            return change.text[0][0];
        }
    }
    highlight_range(range, class_name) {
        return range.editor
            .getDoc()
            .markText(range.start, range.end, { className: class_name });
    }
    apply_edit(workspaceEdit) {
        console.log(workspaceEdit);
        let current_uri = this.connection.getDocumentUri();
        // Specs: documentChanges are preferred over changes
        let changes = workspaceEdit.documentChanges
            ? workspaceEdit.documentChanges.map(change => change)
            : toDocumentChanges(workspaceEdit.changes);
        for (let change of changes) {
            let uri = change.textDocument.uri;
            if (uri !== current_uri) {
                console.warn('Workspace-wide edits not implemented yet');
            }
            else {
                // TODO: show "Renamed X to Y in {change.edits.length} places" in statusbar;
                for (let edit of change.edits) {
                    let start = converter_1.PositionConverter.lsp_to_cm(edit.range.start);
                    let end = converter_1.PositionConverter.lsp_to_cm(edit.range.end);
                    let start_editor = this.virtual_document.get_editor_at_virtual_line(start);
                    let end_editor = this.virtual_document.get_editor_at_virtual_line(end);
                    if (start_editor !== end_editor) {
                        console.log('Edits not implemented for notebooks yet');
                    }
                    else {
                        let doc = start_editor.getDoc();
                        doc.replaceRange(edit.newText, start, end);
                    }
                }
            }
        }
    }
}
/** Return JupyterLab commands to be registered;
 * intended for single-use in index.ts (during extension registration)
 */
CodeMirrorLSPFeature.commands = new Array();
exports.CodeMirrorLSPFeature = CodeMirrorLSPFeature;
