"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const coreutils_1 = require("@jupyterlab/coreutils");
const cm_adapter_1 = require("../codemirror/cm_adapter");
const converter_1 = require("../../converter");
const free_tooltip_1 = require("./components/free_tooltip");
const widgets_1 = require("@phosphor/widgets");
const signaling_1 = require("@phosphor/signaling");
const completion_1 = require("../codemirror/features/completion");
const diagnostics_1 = require("../codemirror/features/diagnostics");
const highlights_1 = require("../codemirror/features/highlights");
const hover_1 = require("../codemirror/features/hover");
const signature_1 = require("../codemirror/features/signature");
const jump_to_1 = require("../codemirror/features/jump_to");
const connection_manager_1 = require("../../connection_manager");
const rename_1 = require("../codemirror/features/rename");
exports.lsp_features = [
    completion_1.Completion,
    diagnostics_1.Diagnostics,
    highlights_1.Highlights,
    hover_1.Hover,
    signature_1.Signature,
    jump_to_1.JumpToDefinition,
    rename_1.Rename
];
/**
 * The values should follow the https://microsoft.github.io/language-server-protocol/specification guidelines
 */
const mime_type_language_map = {
    'text/x-rsrc': 'r',
    'text/x-r-source': 'r',
    // currently there are no LSP servers for IPython we are aware of
    'text/x-ipython': 'python'
};
/**
 * Foreign code: low level adapter is not aware of the presence of foreign languages;
 * it operates on the virtual document and must not attempt to infer the language dependencies
 * as this would make the logic of inspections caching impossible to maintain, thus the WidgetAdapter
 * has to handle that, keeping multiple connections and multiple virtual documents.
 */
class JupyterLabWidgetAdapter {
    constructor(app, widget, rendermime_registry, invoke, server_root) {
        this.app = app;
        this.widget = widget;
        this.rendermime_registry = rendermime_registry;
        this.server_root = server_root;
        this.widget.context.pathChanged.connect(this.reload_connection.bind(this));
        this.invoke_command = invoke;
        this.document_connected = new signaling_1.Signal(this);
        this.adapters = new Map();
        this.connection_manager = new connection_manager_1.DocumentConnectionManager();
        this.connection_manager.closed.connect((manger, { virtual_document }) => {
            console.log('LSP: connection closed, disconnecting adapter', virtual_document.id_path);
            this.disconnect_adapter(virtual_document);
        });
        this.connection_manager.connected.connect((manager, data) => {
            this.on_connected(data).catch(console.warn);
        });
        // register completion connectors
        this.document_connected.connect(() => this.connect_completion());
    }
    get language() {
        // the values should follow https://microsoft.github.io/language-server-protocol/specification guidelines
        if (mime_type_language_map.hasOwnProperty(this.mime_type)) {
            return mime_type_language_map[this.mime_type];
        }
        else {
            let without_parameters = this.mime_type.split(';')[0];
            let [type, subtype] = without_parameters.split('/');
            if (type === 'application' || type === 'text') {
                if (subtype.startsWith('x-')) {
                    return subtype.substr(2);
                }
                else {
                    return subtype;
                }
            }
            else {
                return this.mime_type;
            }
        }
    }
    get root_path() {
        // TODO: serverRoot may need to be included for Hub or Windows, requires testing.
        // let root = PageConfig.getOption('serverRoot');
        return coreutils_1.PathExt.dirname(this.document_path);
    }
    // equivalent to triggering didClose and didOpen, as per syncing specification,
    // but also reloads the connection
    reload_connection() {
        // ignore premature calls (before the editor was initialized)
        if (typeof this.virtual_editor === 'undefined') {
            return;
        }
        // disconnect all existing connections
        this.connection_manager.close_all();
        // recreate virtual document using current path and language
        this.virtual_editor.create_virtual_document();
        // reconnect
        this.connect_document(this.virtual_editor.virtual_document).catch(console.warn);
    }
    invoke_completer(kind) {
        this.current_completion_connector.with_trigger_kind(kind, () => {
            return this.app.commands.execute(this.invoke_command);
        });
    }
    on_connected(data) {
        return __awaiter(this, void 0, void 0, function* () {
            let { virtual_document } = data;
            yield this.connect_adapter(data.virtual_document, data.connection);
            this.document_connected.emit(data);
            yield this.virtual_editor.update_documents().then(() => {
                // refresh the document on the LSP server
                this.document_changed(virtual_document);
                console.log('LSP: virtual document(s) for', this.document_path, 'have been initialized');
            });
        });
    }
    connect_document(virtual_document) {
        return __awaiter(this, void 0, void 0, function* () {
            this.connection_manager.connect_document_signals(virtual_document);
            virtual_document.changed.connect(this.document_changed.bind(this));
            yield this.connect(virtual_document).catch(console.warn);
            virtual_document.foreign_document_opened.connect((host, context) => {
                this.connect(context.foreign_document).catch(console.warn);
            });
        });
    }
    document_changed(virtual_document) {
        // TODO only send the difference, using connection.sendSelectiveChange()
        let connection = this.connection_manager.connections.get(virtual_document.id_path);
        let adapter = this.adapters.get(virtual_document.id_path);
        if (typeof connection === 'undefined') {
            console.log('LSP: Skipping document update signal: connection not ready');
            return;
        }
        if (typeof adapter === 'undefined') {
            console.log('LSP: Skipping document update signal: adapter not ready');
            return;
        }
        console.log('LSP: virtual document', virtual_document.id_path, 'has changed sending update');
        connection.sendFullTextChange(virtual_document.value);
        // guarantee that the virtual editor won't perform an update of the virtual documents while
        // the changes are recorded...
        // TODO this is not ideal - why it solves the problem of some errors,
        //  it introduces an unnecessary delay. A better way could be to invalidate some of the updates when a new one comes in.
        //  but maybe not every one (then the outdated state could be kept for too long fo a user who writes very quickly)
        //  also we would not want to invalidate the updates for the purpose of autocompletion (the trigger characters)
        this.virtual_editor
            .with_update_lock(() => __awaiter(this, void 0, void 0, function* () {
            yield adapter.updateAfterChange();
        }))
            .then()
            .catch(console.warn);
    }
    connect_adapter(virtual_document, connection) {
        return __awaiter(this, void 0, void 0, function* () {
            let adapter = this.create_adapter(virtual_document, connection);
            this.adapters.set(virtual_document.id_path, adapter);
        });
    }
    disconnect_adapter(virtual_document) {
        let adapter = this.adapters.get(virtual_document.id_path);
        this.adapters.delete(virtual_document.id_path);
        if (typeof adapter !== 'undefined') {
            adapter.remove();
        }
    }
    connect(virtual_document) {
        return __awaiter(this, void 0, void 0, function* () {
            let language = virtual_document.language;
            console.log(`LSP: will connect using root path: ${this.root_path} and language: ${language}`);
            let options = {
                virtual_document,
                language,
                root_path: this.root_path,
                server_root: this.server_root,
                document_path: this.document_path
            };
            let connection = this.connection_manager.connect(options).catch(() => {
                this.connection_manager
                    .retry_to_connect(options, 0.5)
                    .catch(console.warn);
            });
            return {
                connection,
                virtual_document
            };
        });
    }
    /**
     * Connect the change signal in order to update all virtual documents after a change.
     *
     * Update to the state of a notebook may be done without a notice on the CodeMirror level,
     * e.g. when a cell is deleted. Therefore a JupyterLab-specific signals are watched instead.
     *
     * While by not using the change event of CodeMirror editors we loose an easy way to send selective,
     * (range) updates this can be still implemented by comparison of before/after states of the
     * virtual documents, which is even more resilient and -obviously - editor-independent.
     */
    connect_contentChanged_signal() {
        this.widget.context.model.contentChanged.connect(this.update_documents.bind(this));
    }
    create_adapter(virtual_document, connection) {
        let adapter_features = new Array();
        for (let feature_type of exports.lsp_features) {
            let feature = new feature_type(this.virtual_editor, virtual_document, connection, this);
            adapter_features.push(feature);
        }
        let adapter = new cm_adapter_1.CodeMirrorAdapter(this.virtual_editor, virtual_document, this, adapter_features);
        console.log('LSP: Adapter for', this.document_path, 'is ready.');
        return adapter;
    }
    update_documents(_slot) {
        // update the virtual documents (sending the updates to LSP is out of scope here)
        this.virtual_editor
            .update_documents()
            .then()
            .catch(console.warn);
    }
    get_position_from_context_menu() {
        // get the first node as it gives the most accurate approximation
        let leaf_node = this.app.contextMenuHitTest(() => true);
        let { left, top } = leaf_node.getBoundingClientRect();
        // @ts-ignore
        let event = this.app._contextMenuEvent;
        // if possible, use more accurate position from the actual event
        // (but this relies on an undocumented and unstable feature)
        if (event !== undefined) {
            left = event.clientX;
            top = event.clientY;
            event.stopPropagation();
        }
        return this.virtual_editor.coordsChar({
            left: left,
            top: top
        }, 'window');
    }
    get_context_from_context_menu() {
        let root_position = this.get_position_from_context_menu();
        let document = this.virtual_editor.document_at_root_position(root_position);
        let connection = this.connection_manager.connections.get(document.id_path);
        let virtual_position = this.virtual_editor.root_position_to_virtual_position(root_position);
        return { document, connection, virtual_position, root_position };
    }
    create_tooltip(markup, cm_editor, position) {
        this.remove_tooltip();
        const bundle = markup.kind === 'plaintext'
            ? { 'text/plain': markup.value }
            : { 'text/markdown': markup.value };
        const tooltip = new free_tooltip_1.FreeTooltip({
            anchor: this.widget.content,
            bundle: bundle,
            editor: this.find_ce_editor(cm_editor),
            rendermime: this.rendermime_registry,
            position: converter_1.PositionConverter.cm_to_ce(position),
            moveToLineEnd: false
        });
        widgets_1.Widget.attach(tooltip, document.body);
        this._tooltip = tooltip;
        return tooltip;
    }
    remove_tooltip() {
        if (this._tooltip !== undefined) {
            this._tooltip.dispose();
        }
    }
}
exports.JupyterLabWidgetAdapter = JupyterLabWidgetAdapter;
