"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const connection_1 = require("./connection");
const signaling_1 = require("@phosphor/signaling");
const coreutils_1 = require("@jupyterlab/coreutils");
const utils_1 = require("./utils");
/**
 * Each Widget with a document (whether file or a notebook) has its own DocumentConnectionManager
 * (see JupyterLabWidgetAdapter), keeping the virtual document spaces separate if a file is opened twice.
 */
class DocumentConnectionManager {
    constructor() {
        this.connections = new Map();
        this.documents = new Map();
        this.ignored_languages = new Set();
        this.connected = new signaling_1.Signal(this);
        this.disconnected = new signaling_1.Signal(this);
        this.closed = new signaling_1.Signal(this);
    }
    connect_document_signals(virtual_document) {
        virtual_document.foreign_document_opened.connect((host, context) => {
            console.log('LSP: Connecting foreign document: ', context.foreign_document.id_path);
            this.connect_document_signals(context.foreign_document);
        });
        virtual_document.foreign_document_closed.connect((host, { foreign_document }) => {
            this.connections.get(foreign_document.id_path).close();
            this.connections.delete(foreign_document.id_path);
            this.documents.delete(foreign_document.id_path);
        });
        this.documents.set(virtual_document.id_path, virtual_document);
    }
    connect_socket(options) {
        let { virtual_document, language, server_root, root_path } = options;
        console.log(root_path);
        // capture just the s?://*
        const wsBase = coreutils_1.PageConfig.getBaseUrl().replace(/^http/, '');
        const wsUrl = `ws${wsBase}lsp/${language}`;
        let socket = new WebSocket(wsUrl);
        let connection = new connection_1.LSPConnection({
            serverUri: 'ws://jupyter-lsp/' + language,
            languageId: language,
            // paths handling needs testing on Windows and with other language servers
            // TODO: compare against: rootUri: 'file:///' + PathExt.join(server_root, root_path),
            rootUri: 'file:///' + coreutils_1.PathExt.join(server_root),
            // TODO: compare against: 'file:///' + PathExt.join(server_root, root_path, virtual_document.uri),
            documentUri: 'file:///' + coreutils_1.PathExt.join(server_root, virtual_document.uri),
            documentText: () => {
                // NOTE: Update is async now and this is not really used, as an alternative method
                // which is compatible with async is used.
                // This should be only used in the initialization step.
                // @ts-ignore
                // if (main_connection.isConnected) {
                //  console.warn('documentText is deprecated for use in JupyterLab LSP');
                // }
                return virtual_document.value;
            }
        }).connect(socket);
        connection.on('error', e => {
            console.warn(e);
            // TODO invalid now
            let error = e.length && e.length >= 1 ? e[0] : new Error();
            // TODO: those codes may be specific to my proxy client, need to investigate
            if (error.message.indexOf('code = 1005') !== -1) {
                console.warn('LSP: Connection failed for ' + virtual_document.id_path);
                console.log('LSP: disconnecting ' + virtual_document.id_path);
                this.closed.emit({ connection, virtual_document });
                this.ignored_languages.add(virtual_document.language);
                console.warn(`Cancelling further attempts to connect ${virtual_document.id_path} and other documents for this language (no support from the server)`);
            }
            else if (error.message.indexOf('code = 1006') !== -1) {
                console.warn('LSP: Connection closed by the server ' + virtual_document.id_path);
            }
            else {
                console.error('LSP: Connection error of ' + virtual_document.id_path + ':', e);
            }
        });
        this.connections.set(virtual_document.id_path, connection);
        return connection;
    }
    retry_to_connect(options, reconnect_delay, retrials_left = -1) {
        return __awaiter(this, void 0, void 0, function* () {
            let { virtual_document } = options;
            if (this.ignored_languages.has(virtual_document.language)) {
                return;
            }
            let interval = reconnect_delay * 1000;
            let success = false;
            while (retrials_left !== 0 && !success) {
                yield this.connect(options)
                    .then(() => {
                    success = true;
                })
                    .catch(e => {
                    console.log(e);
                });
                console.log('LSP: will attempt to re-connect in ' + interval / 1000 + ' seconds');
                yield utils_1.sleep(interval);
                // gradually increase the time delay, up to 5 sec
                interval = interval < 5 * 1000 ? interval + 500 : interval;
            }
        });
    }
    connect(options) {
        return __awaiter(this, void 0, void 0, function* () {
            let connection = this.connect_socket(options);
            let { virtual_document, document_path } = options;
            yield utils_1.until_ready(() => {
                // @ts-ignore
                return connection.isConnected;
            }, 50, 50).catch(() => {
                throw Error('LSP: Connect timed out for ' + virtual_document.id_path);
            });
            console.log('LSP:', document_path, virtual_document.id_path, 'connected.');
            connection.on('close', closed_manually => {
                if (!closed_manually) {
                    console.warn('LSP: Connection unexpectedly disconnected');
                    this.retry_to_connect(options, 0.5).catch(console.warn);
                }
                else {
                    console.warn('LSP: Connection closed');
                    this.closed.emit({ connection, virtual_document });
                }
            });
            this.connected.emit({ connection, virtual_document });
            return connection;
        });
    }
    close_all() {
        for (let [id_path, connection] of this.connections.entries()) {
            let virtual_document = this.documents.get(id_path);
            connection.close();
            // TODO: close() should trigger the closed event, but it does not seem to work, hence manual trigger below:
            this.closed.emit({ connection, virtual_document });
        }
        this.connections.clear();
    }
}
exports.DocumentConnectionManager = DocumentConnectionManager;
